import numpy as np

class BetaPlaneMap:
    def __init__(self, scale, xc, yc, f0, beta, y0=0.):
        """
        scale : map factor from reference mesh (on disk) to physical mesh
        xc, yc : position in physical space of point (0,0) of the reference mesh
        y0 : physical position at which f=f0 : f = f0 + beta*(y-y0) """
        self.scale, self.xc, self.yc = scale, xc, yc
        self.f0, self.beta, self.y0 = f0, beta, y0
    def coriolis(self, x,y): 
        """x,y are coordinates in the physical domain."""
        return self.f0 + self.beta*(y-self.y0)
    def map(self, x, y):
        """x,y are coordinates in the reference domain."""
        return self.xc + self.scale*x, self.yc + self.scale*y
    def map_factor(self, lon,lat): return 0.*lon + self.scale
    def map_angle(self, lon, lat): return 0.*lon

class SphereMap:
    def __init__(self, radius, Omega):
        """
        radius : planetary radius
        Omega : rotation rate => f = 2*Omega*sin(lat)"""
        self.radius, self.Omega = radius, Omega
    def coriolis(self, lon,lat):
       return 2.*self.Omega*np.sin(lat)
    def map(self, lon,lat): return lon, lat
    def map_factor(self, lon,lat): return 0.*lon + self.radius
    def map_angle(self, lon, lat): return 0.*lon

# https://en.wikipedia.org/wiki/Mercator_projection
# https://en.wikipedia.org/wiki/Stereographic_projection
class PolarStereoMap:
    def __init__(self, radius,Omega, delta, phic=0., lonc=0.):
        self.a, self.delta, self.Omega = radius, delta, Omega
        self.phic, self.lonc = phic, lonc # latitude of domain center
    def coriolis(self, lon, lat):
        return 2.*self.Omega*np.sin(lat)
    def map_factor(self,X,Y):
        lon, lat, factor, angle = self.compute_map_factor_angle(X,Y)
        return factor
    def map_angle(self,X,Y):
        lon, lat, factor, angle = self.compute_map_factor_angle(X,Y)
        return angle
    def map(self,X,Y):
        lon, lat, factor, angle = self.compute_map_factor_angle(X,Y)
        return lon, lat
    def compute_map_factor_angle(self,X,Y):
        """ X, Y : reference coordinates on a mesh where cells have unit size"""
        alpha = .5*self.delta/self.a
        X, Y = alpha*X, alpha*Y # scale X and Y to obtain desired resolution delta at domain center
        # now X and Y are as in wikipedia
        # we let (x,y) => (-y,x) to align X axis with equator when phic=0.
        factor =  1./(1.+X*X+Y*Y)
        y = 2.*X*factor
        x = -2.*Y*factor
        z = (1.-X*X-Y*Y)*factor
        # compute d/dX to track mesh orientation
        f2 = factor**2
        dy = 2.*(1.-X*X+Y*Y)*f2
        dx = 4.*X*Y*f2
        dz = -4.*X*f2
        # TODO : rotate x,y,z about polar axis
        x,y,z = self.RotationPhi(x,y,z)
        dx,dy,dz = self.RotationPhi(dx,dy,dz)
        cos_angle = (dy*x-dx*y)/np.sqrt((dx**2+dy**2+dz**2)*(x**2+y**2))
        sin_angle = (dz*(1.-z**2)-z*(dy*y+dx*x))/np.sqrt((dx**2+dy**2+dz**2)*(x**2+y**2))
        angle = np.arctan2(dz*(1.-z**2)-z*(dy*y+dx*x), dy*x-dx*y)
        err = cos_angle-np.cos(angle)
        print 'Consistency check :', err.min(), err.max()
        # TODO : rotate x,y,z so that North Pole gets to desired latitude
        lat = np.arcsin(z)
        lon = np.arctan2(y,x) + self.lonc
        return lon, lat, self.delta*factor, angle
    def RotationPhi(self,x2,y2,z2):
        cos, sin = np.cos(self.phic), np.sin(self.phic)
        x3 = sin * x2 + cos*z2
        z3 = -cos * x2 + sin*z2
        return x3,y2,z3
